@@ -44,12 +44,25 @@ sub access_hash {
     return $self->{$field}->{$key};
 }
 
-sub headers_in {
-    return %{$self->{headers_in}};
+sub access_table {
+    my ($self,$field,$key,$value) = @_;
+    if (wantarray) {
+	return %{$self->{$field}};
+    } elsif (@_==2) {
+	die "FakeApache is unable to simulate the Apache::Table class for the `$field' method called in scalar context.  You'll need to temporarily remove or comment out that method call to use debug files.\n";
+    } elsif (@_==3) {
+	return $self->{$key};
+    } else {
+	return ($self->{$key} = $value);
+    }
 }
 
-sub headers_out {
-    return %{$self->{headers_out}};
+sub args {
+    return (wantarray) ? @{$self->{'args@'}} : $self->{'args$'};
+}
+
+sub get_basic_auth_pw {
+    return @{$self->{get_basic_auth_pw}};
 }
 
 sub cgi_env {
@@ -65,7 +78,20 @@ sub print {
     print (@_);
 }
 
-sub rflush {
+sub write_client {
+    my $self = shift;
+    print (@_);
+}
+
+sub send_fd
+{
+    local (*FD) = @_;
+    print <FD>;
+}
+
+sub warn {
+    my ($self, $msg) = shift;
+    print "warn: $msg\n";
 }
 
 sub send_http_header {
@@ -73,7 +99,7 @@ sub send_http_header {
     $self->content_type('text/plain') if !$self->content_type;
     print "Server: ".$self->cgi_var('SERVER_SOFTWARE')."\n" if $self->cgi_var('SERVER_SOFTWARE');
     print "Content-type: ".$self->content_type."\n";
-    my %headers = %{$self->headers_out};
+    my %headers = %{$self->{headers_out}};
     while (my ($key,$value) = each(%headers)) {
 	print "$key: $value\n";
     }
@@ -90,15 +116,22 @@ sub AUTOLOAD {
     $name =~ s/.*://;   # strip fully-qualified portion
     return if $name eq 'DESTROY';
 
-    if ($name =~ /^(method|method_number|bytes_sent|the_request|proxyreq|header_only|protocol|uri|filename|path_info|content|requires|auth_type|auth_name|document_root|allow_options|content_type|content_encoding|content_language|status|status_line|args)$/) {
+    if ($name =~ /^(allow_options|auth_name|auth_type|bytes_sent|no_cache|content|content_encoding|content_languages|content_type|document_root|filename|header_only|method|method_number|path_info|protocol|proxyreq|requires|status|status_line|the_request|uri)$/) {
 	$self->access_scalar($name,@_);
-    } elsif ($name =~ /^(connection|server)$/) {
+    } elsif ($name =~ /^(as_string|connection|get_remote_host|get_remote_logname|get_server_port|is_initial_req|is_main|server)$/) {
 	$self->read_scalar($name,@_);
-    } elsif ($name =~ /^(header_in|header_out|dir_config|cgi_var)$/) {
-	my $h = {header_in=>'headers_in',header_out=>'headers_out',dir_config=>'dir_config','cgi_env'=>'cgi_env','cgi_var'=>'cgi_env'}->{$name};
+    } elsif ($name =~ /^(dir_config|headers_in|headers_out|err_headers_out|notes|subprocess_env)$/) {
+	$self->access_table($name,@_);	
+    } elsif ($name =~ /^(cgi_var|cgi_env|header_in|header_out|err_header_out|notes)$/) {
+	my $h = {header_in=>'headers_in',header_out=>'headers_out',err_header_out=>'err_headers_out','cgi_env'=>'cgi_env','cgi_var'=>'cgi_env'}->{$name};
 	$self->access_hash($h,@_);
+    } elsif ($name =~ /^(log_error|log_reason|post_connection|note_basic_auth_failure|register_cleanup|send_cgi_header|custom_response|rflush|cgi_header_out)$/) {
+	# do nothing and hope it's okay
+    } elsif ($name =~ /^(last|soft_timeout|kill_timeout|lookup_file|lookup_uri|main|prev|next|handler|hard_timeout|reset_timeout|internal_redirect|internal_redirect_handler)$/) {
+	die "FakeApache is unable to simulate the `$name' method.  You'll need to temporarily remove or comment out that method call to use debug files.\n";
     } else {
-        die "No such function `$name' in class $type";
+        CORE::warn("FakeApache does not know about the `$name' method; returning undef.\n");
+	return undef;
     }
 }
 
@@ -129,8 +162,11 @@ sub AUTOLOAD {
 
     if ($name =~ /^(server_admin|server_hostname|port|is_virtual|names)$/) {
 	$self->access_scalar($name,@_);
+    } elsif ($name =~ /^(warn|log_error)$/) {
+	# do nothing and hope it's okay
     } else {
-        die "No such function `$name' in class $type";
+        warn "FakeApache::Server does not know about the `$name' method; returning undef.\n";
+	return undef;
     }
 }
 
@@ -162,7 +198,8 @@ sub AUTOLOAD {
     if ($name =~ /^(remote_host|remote_ip|local_addr|remote_addr|remote_logname|user|auth_type|aborted)$/) {
 	$self->access_scalar($name,@_);
     } else {
-        die "No such function `$name' in class $type";
+        warn "FakeApache::Connection does not know about the `$name' method; returning undef.\n";
+	return undef;
     }
 }
 
